"""Jobs manager client.

Copyright 2020 Caliber Data Labs
"""

#  Copyright (c) 2020 Caliber Data Labs.
#  All rights reserved.
#

import json
import time
from typing import List, NamedTuple, Optional, Union, Callable

import requests

from ..constants import ACK_OK_RESPONSE, JobConstants, JobTags, JobTypes
from ..custom_exceptions import JobError
from ..tools.logger import get_logger
from ..tools.misc import join, retry

DEFAULT_SEARCH_SIZE = 500

# seconds. Note that this determines the minimum job time.
CHECK_JOB_STATUS_INTERVAL = 10

UPDATE_JOB_PROGRESS_STEP = 5

logger = get_logger(__name__)

STATUS_INTERVAL = 30


class Job(NamedTuple):
    id: str
    status: str
    progress: float
    description: str
    stage: str = ""

class JobManagerClient(JobConstants):
    """Client for interacting with the Job Manager."""

    def __init__(
        self,
        asset_manager_url: str,
        auth_token: str,
        job_id: Optional[str] = None,
        asset_type: Optional[str] = None,
        asset_id: Optional[str] = None,
        username: Optional[str] = None,
        job_type: Optional[str] = None,
        notes: Optional[Union[List[str], str]] = None,
        description: Optional[str] = None,
        stage: Optional[str] = None,
        tags: Optional[List[str]] = None,
        user_facing: bool = False,
        task_sync: bool = False
    ):
        """
        asset_manager_url: URL to asset manager
        auth_token: Auth token to use
        job_id: unique identifier of a job
        asset_type: Type of asset ( used for updating job metadata)
        asset_id: Asset ID (used for updating job metadata)
        username: Username (used for updating job metadata)
        job_type: Type of the job (used for updating job metadata)
        notes: time-history of the events in the job (started at what time, queued at what time) (used for updating job metadata)
        description: the description of the job (used for updating job metadata)
        stage: Job stage (used for updating job metadata)
        ttl: job's time to live in seconds
        sync_interval: Time interval to wait before checking if timeout > TTL or job failure,
        """

        self.jobs_api_url = join(asset_manager_url, "jobs")
        self.auth_token = auth_token

        if not job_id:
            tags = tags or []
            if user_facing:
                tags.append(JobTags.USER_FACING)

            job_id = self.create_job(
                asset_type,
                asset_id,
                username,
                job_type,
                notes,
                description,
                stage,
                tags,
                task_sync
            )

        self.job_id = job_id
        self.last_progress = 0.0
        self.last_stage = None

    def create_job(
            self,
            asset_type: str,
            asset_id: str,
            username: str,
            job_type: str,
            notes: list,
            description: str,
            stage: str,
            tags: List[str],
            task_sync: bool
    ) -> str:

        if job_type not in JobTypes.VALID_TYPES:
            raise JobError("invalid job type {}".format(job_type))

        asset_id = asset_id or 'undefined'

        json_payload = {
            "asset_type": asset_type,
            "asset_id": asset_id,
            "user": username,
            "job_type": job_type or '',
            "notes": notes or [],
            "description": description or '',
            "stage": stage or '',
            "tags": tags or [],
            "task_sync": task_sync
        }
        params = self.add_asset_manager_params()

        url = self.get_jobs_url("add/")
        resp = requests.post(url, json=json_payload, params=params)
        if not resp.ok:
            raise JobError(
                "Could not create the job ::: {} -- {}".format(
                    resp.reason, resp.content
                )
            )
        return resp.json()["job_id"]

    def queue(self):
        url = self.get_jobs_url('queue/{}'.format(self.job_id))
        params = self.add_asset_manager_params()
        resp = requests.put(url, params=params)
        if not resp.ok:
            raise JobError(
                "Could not queue the job ::: {} -- {}".format(
                    resp.reason, resp.content)
            )
        return resp.json()

    def run(self,
            cpu_core: Optional[int] = None,
            hostname: Optional[str] = None):

        url = self.get_jobs_url('run/{}'.format(self.job_id))

        json_payload = {}
        if cpu_core is not None:
            json_payload["cpu_core"] = cpu_core
        if hostname is not None:
            json_payload["hostname"] = hostname

        params = self.add_asset_manager_params()
        resp = requests.put(url, params=params, json=json_payload)
        if not resp.ok:
            raise JobError(
                "Could not run the job ::: {} -- {}".format(
                    resp.reason, resp.content)
            )
        return resp.json()

    def finish(self):
        if self.get_status() == JobConstants.STATUS_FINISHED:
            return
        url = self.get_jobs_url('finish/{}'.format(self.job_id))
        params = self.add_asset_manager_params()
        resp = requests.put(url, params=params)
        if not resp.ok:
            raise JobError(
                "Could not finish the job ::: {} -- {}".format(
                    resp.reason, resp.content
                )
            )
        return resp.json()

    def abort(self,
              revoke_celery_job: bool = False,
              error_string: str = '',
              exception: Optional[Exception] = None):
        if exception is not None:
            error_string = repr(exception)
        else:
            error_string = str(error_string)[0:1024]
        self.update_stage(stage="Error :{}".format(error_string))
        if self.get_status() == JobConstants.STATUS_ABORTED:
            return
        url = self.get_jobs_url('abort/{}'.format(self.job_id))
        params = {'error_string': error_string,
                  'revoke': revoke_celery_job}
        params = self.add_asset_manager_params(params)
        resp = requests.put(url, params=params)
        if not resp.ok:
            raise JobError(
                "Could not abort the job ::: {} -- {}".format(
                    resp.reason, resp.content)
            )
        return resp.json()

    def append_note(self, note):
        url = self.get_jobs_url('append_note/{}'.format(self.job_id))
        params = {'note': note}
        params = self.add_asset_manager_params(params)
        resp = requests.put(url, params=params)
        if not resp.ok:
            raise JobError(
                "Could not append_note to the job ::: {} -- {}".format(
                    resp.reason, resp.content
                )
            )
        return resp.json()

    def update_meta(self, key, value):
        url = self.get_jobs_url('meta/{}'.format(self.job_id))
        data = {'key': key, 'value': value}
        params = self.add_asset_manager_params()
        resp = requests.put(url, json=data, params=params)
        if not resp.ok:
            raise JobError(
                "Could not update meta of the job ::: {} -- {}".format(
                    resp.reason, resp.content
                )
            )
        return resp.json()

    def get_meta(self):
        url = self.get_jobs_url('meta/{}'.format(self.job_id))
        params = self.add_asset_manager_params()
        resp = requests.get(url, params=params)
        if not resp.ok:
            raise JobError(
                "Could not get meta of the job ::: {} -- {}".format(
                    resp.reason, resp.content
                )
            )
        return resp.json().get("dict", {})

    def update_stage(self, stage):
        url = self.get_jobs_url('stage/{}'.format(self.job_id))
        params = {'stage': stage}
        params = self.add_asset_manager_params(params)
        resp = requests.put(url, params=params)
        if not resp.ok:
            raise JobError(
                "Could not update stage of the job ::: {} -- {}".format(
                    resp.reason, resp.content
                )
            )
        return resp.json()

    @retry(Exception, logger=logger, initial_wait=1)
    def get_status(self) -> str:
        url = self.get_jobs_url('status/{}'.format(self.job_id))
        params = self.add_asset_manager_params()
        resp = requests.get(url, params=params)
        if not resp.ok:
            raise JobError(
                "Could not get status of the job ::: {} -- {} -- with auth token: {}".format(
                    resp.reason, resp.content, self.auth_token))
        return resp.json()["status"]

    @retry(Exception, logger=logger, initial_wait=1)
    def get_comprehensive_status(self) -> dict:
        url = self.get_jobs_url('comprehensive_status/{}'.format(self.job_id))
        params = self.add_asset_manager_params()
        resp = requests.get(url, params=params)
        if not resp.ok:
            raise JobError(
                "Could not get status of the job ::: {} -- {} -- with auth token: {}".format(
                    resp.reason, resp.content, self.auth_token))
        return resp.json()

    @retry(Exception, logger=logger, initial_wait=1)
    def get_progress(self) -> float:
        url = self.get_jobs_url('progress/{}'.format(self.job_id))
        params = self.add_asset_manager_params()
        resp = requests.get(url, params=params)
        if not resp.ok:
            raise JobError("Could not get the job progress")
        if not resp.ok:
            raise JobError(
                "Could not get progress the job ::: {} -- {}".format(
                    resp.reason, resp.content
                )
            )
        return resp.json()["progress"]

    def update_progress(self, progress: float, raise_on_abort_canceled: bool = False,
                        message: str = None, force_update=False):
        """
        Args:
            progress (float): percentage of job progress
            raise_on_abort_canceled (bool): raise JobError if job is canceled or aborted when this is True
            progress_message: status update string
        """
        if raise_on_abort_canceled:
            job_status = self.get_status()
            if job_status in {JobConstants.STATUS_ABORTED}:
                raise JobError('Job is {}'.format(job_status))

        if (progress - self.last_progress) > UPDATE_JOB_PROGRESS_STEP or progress >= 100.0\
                or (self.last_stage != message) or force_update:
            self.last_progress = progress
            params = {'progress': round(progress, 2)}
            if message is not None:
                self.update_stage(message)

            url = self.get_jobs_url('progress/{}'.format(self.job_id))
            params = self.add_asset_manager_params(params)
            resp = requests.put(url, params=params)
            if not resp.ok:
                logger.error(
                    "Could not update_progress of the job ::: {} -- {}".format(
                        resp.reason, resp.content
                    )
                )
            return resp.json()

    def set_resubmit_data(self, resubmission_dict: dict):
        """Set the data for job resubmission."""
        resubmission_str = json.dumps(resubmission_dict)
        self.update_meta('resubmit', resubmission_str)
        return ACK_OK_RESPONSE

    def resubmit(self):
        """Resubmit an extraction job."""
        url = self.get_jobs_url('resubmit/{}'.format(self.job_id))
        params = self.add_asset_manager_params()
        resp = requests.put(url, params=params)
        if not resp.ok:
            raise JobError(
                "Could not resubmit the job ::: {} -- {}".format(
                    resp.reason, resp.content
                )
            )
        return resp.json()

    def get_jobs_url(self, route):
        url = join(self.jobs_api_url, route)
        return url

    def add_asset_manager_params(self, params_dict=None):
        if params_dict is None:
            params_dict = {}
        if self.auth_token:
            params_dict["token"] = self.auth_token
        return params_dict

    def add_child_job(self, job_id: str):
        self.update_meta(key="child_job_ids", value=[job_id])

    # def get_task_status(self,task_id: str):
    #     url = self.get_jobs_url(f"query_task/{task_id}")
    #     params = self.add_asset_manager_params()
    #     return requests.get(url, params=params).json()

    def abort_task(self,task_id: str):
        url = self.get_jobs_url(f"query_task/{task_id}/abort")
        params = self.add_asset_manager_params()
        return requests.get(url, params=params).json()


def wait_for_jobs(
        job_ids: List[str],
        asset_manager_url: str,
        auth_token: str,
        timeout: Optional[int] = None,
        raise_on_error: bool = False,
        progress_callback: Optional[Callable] = None,
):
    failed_enrichment_jobs = set()
    successful_job_ids = set()
    logger.info("Waiting for {} jobs to finish:: {}".format(
        len(job_ids), job_ids[0:10]))

    # Wait for all enrichment tasks to finish
    job_ids_to_clients = dict(
        [
            (
                job_id,
                JobManagerClient(
                    job_id=job_id,
                    asset_manager_url=asset_manager_url,
                    auth_token=auth_token
                )
            )
            for job_id in job_ids
        ]
    )
    logger.info("jobs being done: {}".format(
        job_ids_to_clients))
    t1 = time.time()
    finished_jobs = set()
    job_id_to_progress_dict = {}

    while True:
        for job_id, job_manager_client in job_ids_to_clients.items():
            job_status = job_manager_client.get_status()
            job_id_to_progress_dict[job_id] = job_manager_client.get_progress()

            if job_status in {
                JobManagerClient.STATUS_FINISHED,
                JobManagerClient.STATUS_ABORTED}:
                if job_status == JobManagerClient.STATUS_FINISHED:
                    successful_job_ids.add(job_id)
                    logger.info('Job {} finished'.format(job_id))
                elif job_status == JobManagerClient.STATUS_ABORTED:
                    failed_enrichment_jobs.add(job_id)
                    logger.info('Job {} failed'.format(job_id))
                    if raise_on_error:
                        raise JobError(
                            "Job {} encountered failure".format(job_id))
                finished_jobs.add(job_id)

        # Update callback with progress
        if progress_callback:
            progress_callback(job_id_to_progress_dict)

        job_ids_to_clients = {
            k: v for k,
                     v in job_ids_to_clients.items() if k not in finished_jobs}

        if len(job_ids_to_clients) == 0:
            break
        else:
            logger.info(
                "Waiting for {} jobs to finish:: {}".format(
                    len(job_ids_to_clients), list(
                        job_ids_to_clients.keys())[
                                             0:5]))

            time.sleep(CHECK_JOB_STATUS_INTERVAL)
            if timeout and time.time() - t1 > timeout:
                raise TimeoutError(
                    "Waiting for {} jobs exceeded timeout of {}s".format(
                        len(job_ids), timeout))

    return {
        "successful": len(list(successful_job_ids)),
        "failed": len(list(failed_enrichment_jobs))
    }



from concurrent.futures import Future, as_completed
from typing import Dict


def create_exception_map(
        future_map: Dict[Future, str]) -> Dict[str, Exception]:
    id_exception_map = {}
    for future in as_completed(future_map):
        id = future_map[future]
        try:
            future.result()
        except Exception as exc:
            id_exception_map[id] = exc
    return id_exception_map



"""Mock the fetching of file annotations from Deepen."""


#  Copyright (c) 2020 Caliber Data Labs.
#  All rights reserved.
#


def get_file_annotations(client_id, dataset_id, file_id):
    return {
        "labels": [
            {
                "client_id": client_id,
                "dataset_id": dataset_id,
                "file_id": file_id,
                "label_category_id": "car",
                "label_id": "car:6",
                "label_type": "box",
                "project_id": "2mjs8n87",
                "stage_id": "QA",
                "attributes": {},
                "box": [446, 215, 273, 211],
                "create_time_millis": 1568722395611,
                "labeller_email": "auto_interpolation",
                "update_time_millis": 1568722395611,
            },
            {
                "client_id": client_id,
                "dataset_id": dataset_id,
                "file_id": file_id,
                "label_category_id": "car",
                "label_id": "car:7",
                "label_type": "box",
                "project_id": "2mjs8n87",
                "stage_id": "QA",
                "attributes": {},
                "box": [263, 210, 235, 136],
                "create_time_millis": 1568722395611,
                "labeller_email": "auto_interpolation",
                "update_time_millis": 1568722395611,
            },
            {
                "client_id": client_id,
                "dataset_id": dataset_id,
                "file_id": file_id,
                "label_category_id": "car",
                "label_id": "car:8",
                "label_type": "box",
                "project_id": "2mjs8n87",
                "stage_id": "QA",
                "attributes": {},
                "box": [550, 201, 130, 98],
                "create_time_millis": 1568722395611,
                "labeller_email": "auto_interpolation",
                "update_time_millis": 1568722395611,
            },
            {
                "client_id": client_id,
                "dataset_id": dataset_id,
                "file_id": file_id,
                "label_category_id": "car",
                "label_id": "car:9",
                "label_type": "box",
                "project_id": "2mjs8n87",
                "stage_id": "QA",
                "attributes": {},
                "box": [424, 200, 109, 63],
                "create_time_millis": 1568722395611,
                "labeller_email": "auto_interpolation",
                "update_time_millis": 1568722395611,
            },
            {
                "client_id": client_id,
                "dataset_id": dataset_id,
                "file_id": file_id,
                "label_category_id": "car",
                "label_id": "car:10",
                "label_type": "box",
                "project_id": "2mjs8n87",
                "stage_id": "QA",
                "attributes": {},
                "box": [589, 185, 82, 70],
                "create_time_millis": 1568722395611,
                "labeller_email": "auto_interpolation",
                "update_time_millis": 1568722395611,
            },
            {
                "client_id": client_id,
                "dataset_id": dataset_id,
                "file_id": file_id,
                "label_category_id": "adult",
                "label_id": "adult:15",
                "label_type": "box",
                "project_id": "2mjs8n87",
                "stage_id": "QA",
                "attributes": {},
                "box": [1046, 201, 154, 205],
                "create_time_millis": 1568722395611,
                "labeller_email": "auto_interpolation",
                "update_time_millis": 1568722395611,
            },
            {
                "client_id": client_id,
                "dataset_id": dataset_id,
                "file_id": file_id,
                "label_category_id": "personal mobility",
                "label_id": "personal mobility:16",
                "label_type": "box",
                "project_id": "2mjs8n87",
                "stage_id": "QA",
                "attributes": {},
                "box": [889, 230, 152, 130],
                "create_time_millis": 1568722395611,
                "labeller_email": "auto_interpolation",
                "update_time_millis": 1568722395611,
            },
            {
                "client_id": client_id,
                "dataset_id": dataset_id,
                "file_id": file_id,
                "label_category_id": "car",
                "label_id": "car:18",
                "label_type": "box",
                "project_id": "2mjs8n87",
                "stage_id": "QA",
                "attributes": {},
                "box": [321, 191, 85, 46],
                "create_time_millis": 1568722395611,
                "labeller_email": "auto_interpolation",
                "update_time_millis": 1568722395611,
            },
            {
                "client_id": client_id,
                "dataset_id": dataset_id,
                "file_id": file_id,
                "label_category_id": "rigid",
                "label_id": "rigid:19",
                "label_type": "box",
                "project_id": "2mjs8n87",
                "stage_id": "QA",
                "attributes": {},
                "box": [466, 129, 108, 112],
                "create_time_millis": 1568722395611,
                "labeller_email": "auto_interpolation",
                "update_time_millis": 1568722395611,
            },
            {
                "client_id": client_id,
                "dataset_id": dataset_id,
                "file_id": file_id,
                "label_category_id": "car",
                "label_id": "car:20",
                "label_type": "box",
                "project_id": "2mjs8n87",
                "stage_id": "QA",
                "attributes": {},
                "box": [1050, 140, 230, 133],
                "create_time_millis": 1568722395611,
                "labeller_email": "auto_interpolation",
                "update_time_millis": 1568722395611,
            },
            {
                "client_id": client_id,
                "dataset_id": dataset_id,
                "file_id": file_id,
                "label_category_id": "car",
                "label_id": "car:21",
                "label_type": "box",
                "project_id": "2mjs8n87",
                "stage_id": "QA",
                "attributes": {},
                "box": [889, 154, 228, 80],
                "create_time_millis": 1568722395611,
                "labeller_email": "auto_interpolation",
                "update_time_millis": 1568722395611,
            },
            {
                "client_id": client_id,
                "dataset_id": dataset_id,
                "file_id": file_id,
                "label_category_id": "car",
                "label_id": "car:22",
                "label_type": "box",
                "project_id": "2mjs8n87",
                "stage_id": "QA",
                "attributes": {},
                "box": [852, 155, 211, 72],
                "create_time_millis": 1568722395611,
                "labeller_email": "auto_interpolation",
                "update_time_millis": 1568722395611,
            },
            {
                "client_id": client_id,
                "dataset_id": dataset_id,
                "file_id": file_id,
                "label_category_id": "car",
                "label_id": "car:23",
                "label_type": "box",
                "project_id": "2mjs8n87",
                "stage_id": "QA",
                "attributes": {},
                "box": [766, 162, 119, 48],
                "create_time_millis": 1568722395611,
                "labeller_email": "auto_interpolation",
                "update_time_millis": 1568722395611,
            },
            {
                "client_id": client_id,
                "dataset_id": dataset_id,
                "file_id": file_id,
                "label_category_id": "car",
                "label_id": "car:24",
                "label_type": "box",
                "project_id": "2mjs8n87",
                "stage_id": "QA",
                "attributes": {},
                "box": [1100, 139, 164, 61],
                "create_time_millis": 1568722395611,
                "labeller_email": "auto_interpolation",
                "update_time_millis": 1568722395611,
            },
            {
                "client_id": client_id,
                "dataset_id": dataset_id,
                "file_id": file_id,
                "label_category_id": "car",
                "label_id": "car:25",
                "label_type": "box",
                "project_id": "2mjs8n87",
                "stage_id": "QA",
                "attributes": {},
                "box": [1174, 139, 106, 60],
                "create_time_millis": 1568722395611,
                "labeller_email": "auto_interpolation",
                "update_time_millis": 1568722395611,
            },
            {
                "client_id": client_id,
                "dataset_id": dataset_id,
                "file_id": file_id,
                "label_category_id": "car",
                "label_id": "car:26",
                "label_type": "box",
                "project_id": "2mjs8n87",
                "stage_id": "QA",
                "attributes": {},
                "box": [1244, 138, 36, 60],
                "create_time_millis": 1568722395611,
                "labeller_email": "auto_interpolation",
                "update_time_millis": 1568722395611,
            },
            {
                "client_id": client_id,
                "dataset_id": dataset_id,
                "file_id": file_id,
                "label_category_id": "car",
                "label_id": "car:32",
                "label_type": "box",
                "project_id": "2mjs8n87",
                "stage_id": "QA",
                "attributes": {},
                "box": [783, 162, 145, 49],
                "create_time_millis": 1568722395611,
                "labeller_email": "auto_interpolation",
                "update_time_millis": 1568722395611,
            },
            {
                "client_id": client_id,
                "dataset_id": dataset_id,
                "file_id": file_id,
                "label_category_id": "car",
                "label_id": "car:35",
                "label_type": "box",
                "project_id": "2mjs8n87",
                "stage_id": "QA",
                "attributes": {},
                "box": [1066, 149, 161, 49],
                "create_time_millis": 1568722395611,
                "labeller_email": "auto_interpolation",
                "update_time_millis": 1568722395611,
            },
            {
                "client_id": client_id,
                "dataset_id": dataset_id,
                "file_id": file_id,
                "label_category_id": "car",
                "label_id": "car:36",
                "label_type": "box",
                "project_id": "2mjs8n87",
                "stage_id": "QA",
                "attributes": {},
                "box": [918, 159, 101, 34],
                "create_time_millis": 1568722395611,
                "labeller_email": "auto_interpolation",
                "update_time_millis": 1568722395611,
            },
            {
                "client_id": client_id,
                "dataset_id": dataset_id,
                "file_id": file_id,
                "label_category_id": "car",
                "label_id": "car:37",
                "label_type": "box",
                "project_id": "2mjs8n87",
                "stage_id": "QA",
                "attributes": {},
                "box": [885, 162, 87, 29],
                "create_time_millis": 1568722395611,
                "labeller_email": "auto_interpolation",
                "update_time_millis": 1568722395611,
            },
        ]
    }



import io
from typing import NamedTuple, Optional, Union, List
from datetime import datetime

from ..models.annotation import Annotation


class Image(NamedTuple):
    image_path: Optional[str] = None
    id: Optional[str] = None
    image_url: Optional[str] = None
    image_uri: Optional[str] = None
    image_bytes: Optional[Union[io.BytesIO, bytes, str]] = None
    sensor: Optional[str] = None
    timestamp: Optional[Union[str, datetime]] = None
    session_uid: Optional[str] = None
    annotations: Optional[List[Annotation]] = None
    aux_metadata: Optional[dict] = None



"""
To create a whl file run python3 setup.py sdist bdist_wheel in scenebox_clients_packaging directory
The whl file (scenebox-0.0.1-py3-none-any.whl) can be used to run the test in helper.py
"""
import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()
with open("requirements.txt", "r") as file:
    INSTALL_REQS = file.read().split("\n")

with open('../BASE_VERSION') as version_file:
    version = version_file.readline()

print(INSTALL_REQS)
setuptools.setup(
    name="scenebox",
    version=version,
    author="Caliber Data labs",
    author_email="yaser@caliberdatalabs.ai",
    description="Caliber Data Labs Clients package",
    long_description=long_description,
    long_description_content_type="text/markdown",
    # url="https://github.com/pypa/sampleproject",
    packages=setuptools.find_packages(),
    package_data={'': ['BASE_VERSION']},
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    install_requires=INSTALL_REQS
)


#  Copyright (c) 2021 Caliber Data Labs.
#  All rights reserved.
#
from typing import NamedTuple, Optional
import requests
from PIL import Image, ImageFile
from ..custom_exceptions import ResponseError

from .logger import get_logger

logger = get_logger(__name__)


class ImageProperties(NamedTuple):
    width: int
    height: int
    format: str


def get_image_attributes(image_url: Optional[str] = None,
                         image_path: Optional[str] = None,
                         image_bytes: Optional[bytes] = None):
    # get file size *and* image size (None if not known)
    if image_url:
        # TODO is this safe to do?
        resp = requests.get(image_url, stream=True)
        if not resp.ok:
            raise ResponseError(
                "Could not find an image in {}".format(image_url))
        else:
            p = ImageFile.Parser()
            for data in resp.iter_content(chunk_size=1024):
                p.feed(data)
                if p.image:
                    return ImageProperties(
                        width=p.image.width,
                        height=p.image.height,
                        format=p.image.format
                    )

    elif image_path:
        with Image.open(image_path) as image_file:
            width, height = image_file.size
            format = image_file.format
            return ImageProperties(
                width=width,
                height=height,
                format=format)
    elif image_bytes:
        image = Image.open(image_bytes)
        width, height = image.size
        format = image.format
        return ImageProperties(
            width=width,
            height=height,
            format=format)
    else:
        raise NotImplementedError(
            "Either image url or image path must be provided")



#  Copyright (c) 2020 Caliber Data Labs.
#  All rights reserved.
#

import json
import logging

from scene_engine.rest import ASSET_MANAGER_URL
from shared.external.clients.asset_manager_client import AssetManagerClient
from shared.external.constants import (AnnotationMediaTypes,
                                       AnnotationProviders, AnnotationTypes,
                                       AssetsConstants, AnnotationGroups)
from shared.external.custom_exceptions import InvalidAnnotationError
from shared.external.models.annotation import Annotation, AnnotationEntity
from shared.external.tools import time_utils

annotation_projects_asset_manager = AssetManagerClient(
    AssetsConstants.ANNOTATION_PROJECTS_ASSET_ID,
    metadata_only=True,
    asset_manager_url=ASSET_MANAGER_URL,
)


class DeepenAnnotation(Annotation):
    def __init__(
        self,
        annotation_project_id,
        file_id,
        file_labels,
        resolution=None,
        annotation_binary_id=None,
        version=None
    ):
        annotation_meta = annotation_projects_asset_manager.get_metadata(
            id=annotation_project_id
        )
        valid = False
        for set_id, set_files in json.loads(annotation_meta["state"]).items():
            if file_id in set_files:
                valid = True
                break
        if not valid:
            raise ValueError(
                "The file ID {} does not exist in the Annotation {}".format(
                    file_id, annotation_project_id
                )
            )
        self.file_id = file_id
        self.resolution = resolution
        self.annotation_meta = annotation_meta
        self.annotation_meta["binary_id"] = annotation_binary_id
        # Always only one set
        self.annotation_set_id = annotation_meta["sets"][0]
        self.file_labels = file_labels
        self.annotation_entities = []
        self._version = version
        self.timestamp = time_utils.string_to_datetime(
            self.annotation_meta.get("timestamp")
        )
        super().__init__(annotation_meta=annotation_meta)

    def set_id(self):
        self.id = self.annotation_meta["id"] + self.file_id + ".ann"

    def set_ground_truth(self):
        self.annotation_group = AnnotationGroups.GROUND_TRUTH

    def set_version(self):
        self.version = self._version

    def set_media_type(self):
        # TODO! FIXME! : Handle other annotation types?
        self.media_type = AssetsConstants.IMAGES_ASSET_ID

    def set_provider(self):
        self.provider = AnnotationProviders.DEEPEN

    def set_annotation_type(self):
        self.annotation_type = AnnotationTypes.TWO_D_BOUNDING_BOX

    def set_annotation_entities(self):
        # In general, there can be multiple types of labels for a single file
        # Never exit on error
        for deepen_label in self.file_labels:
            try:
                label_type = deepen_label["label_type"]
                if label_type == "box":
                    self.set_bounding_box_annotation_entities(deepen_label)
                elif label_type == "polygon":
                    self.set_polygon_annotation_entities(deepen_label)
            except Exception as e:
                logging.error(
                    "Could not set annotation entity::: {}".format(e))

    def set_set_membership(self):
        self.sets = [self.annotation_set_id]

    def set_bounding_box_annotation_entities(self, deepen_label):
        """
        Note: box is an array of 4 numbers representing a 2D bounding box. The first two numbers are the x, y
        coordinates of the top left corner of the box. 3rd and 4th numbers represent length of box
        along x and y axes. Coordinate system has origin at the top left corner of the image.
        """
        box = deepen_label["box"]
        try:
            left = box[0]
            top = box[1]
            width = box[2]
            height = box[3]
            label = deepen_label["label_category_id"]
        except KeyError as e:
            raise InvalidAnnotationError(str(e))
        key = "{}_{}".format(deepen_label["file_id"], deepen_label["label_id"])
        coordinates = [
            {"x": left, "y": top},
            {"x": left + width, "y": top},
            {"x": left + width, "y": top + height},
            {"x": left, "y": top + height},
        ]
        self.annotation_entities.append(
            AnnotationEntity(
                label=label,
                annotation_type=AnnotationTypes.TWO_D_BOUNDING_BOX,
                coordinates=coordinates,
                attributes=deepen_label.get("attributes")
            )
        )

    def set_polygon_annotation_entities(self, deepen_label):
        """
        Note: Array of polygons. Each polygon is an array of points [x, y].
        """
        for polygon_part, polygon in enumerate(deepen_label["polygons"]):
            try:
                coordinates = [{"x": i[0], "y": i[1]} for i in polygon]
                key = "{}_{}_{}".format(
                    deepen_label["file_id"],
                    deepen_label["label_id"],
                    polygon_part)
                label = deepen_label["label_category_id"]
                coordinates.append(coordinates[0])
            except KeyError as e:
                raise InvalidAnnotationError(str(e))
            self.annotation_entities.append(
                AnnotationEntity(
                    label=label,
                    annotation_type=AnnotationTypes.POLYGON,
                    coordinates=coordinates,
                    attributes=deepen_label.get("attributes"),
                    uid=key
                )
            )
